! dc_string.f90 - string module for Fortran90
! Copyright (C) TOYODA Eizi, 2000.  All rights reserved.
! vi: set ts=8 sw=4:

! 
!	{W[͉ϒێ\̂񋟂B
!
!	{W[ ISO/IEC 1539-2:1994 ŋK肳Ă
!	ϒ񃂃W[Ƃ̊֌W͖łB

module dc_string

    implicit none

    private
    public STRING, len, var_str, char, get, put, put_line, &
	& assignment(=), operator(==), operator(/=), operator(//), &
	& operator(<), operator(<=), operator(>), operator(>=), &
	& STRING_TABLE_ENTRY, disposable, string_table

    type STRING
	integer::				id
    end type

    ! Iɕێ镶\̌^
    type STRING_TABLE_ENTRY
	! nlink: ɂ 0 ɂȂB
	! l𐶐Ƃɂ͑ȂB
	! VARYING_STRING ϐɊi[Ƃ 1 A
	! ϐsvɂȂƂ 1 B
	! 1  0 ɂȂƂ buffer pׂłB
    	integer::			nlink
	! p̃^C~Ov
	integer::			serial
	! ɂċԂɂȂ
	character, pointer::		buffer(:) 
    end type

    interface disposable
	module procedure string_set_disposable
    end interface

    interface len
	module procedure string_len
    end interface

    interface var_str
	module procedure char_to_string
    end interface

    interface char
	module procedure string_to_char_all
	module procedure string_to_char_length
    end interface

    interface get
	module procedure string_get
	module procedure string_get_default
    end interface

    interface put
	module procedure string_put
	module procedure string_put_default
	module procedure char_put
	module procedure char_put_default
    end interface

    interface put_line
	module procedure string_put_line
	module procedure string_put_line_default
	module procedure char_put_line
	module procedure char_put_line_default
    end interface

    interface scan
	module procedure string_scan_string
	module procedure string_scan_char
    end interface

    interface verify
	module procedure string_verify_string
	module procedure string_verify_char
    end interface

    interface index
	module procedure string_index_string
	module procedure string_index_char
	module procedure char_index_string
    end interface

    interface assignment(=)
	module procedure string_let_char
	module procedure char_let_string
	module procedure string_let_string
    end interface

    interface operator(//)
	module procedure string_add_string
	module procedure char_add_string
	module procedure string_add_char
    end interface

    interface operator(==)
	module procedure string_eq_string
	module procedure string_eq_char
	module procedure char_eq_string
    end interface

    interface operator(/=)
	module procedure string_ne_string
	module procedure string_ne_char
	module procedure char_ne_string
    end interface

    interface operator(<)
	module procedure string_lt_string
	module procedure string_lt_char
	module procedure char_lt_string
    end interface

    interface operator(<=)
	module procedure string_le_string
	module procedure string_le_char
	module procedure char_le_string
    end interface

    interface operator(>)
	module procedure string_gt_string
	module procedure string_gt_char
	module procedure char_gt_string
    end interface

    interface operator(>=)
	module procedure string_ge_string
	module procedure string_ge_char
	module procedure char_ge_string
    end interface

    !
    ! rẐ߂̃V{
    !

    integer, parameter::	CMP_LESSTHAN = -1
    integer, parameter::	CMP_EQUIVALENT = 0
    integer, parameter::	CMP_GREATERTHAN = 1
    integer, parameter::	CMP_ERROR = -100

    ! --- \ ---
    !
    !  STRING_TABLE_ENTRY \̂Ȃ string_table Ɋi[B
    ! string_table  ɏ߂ĊmۂƂɂ͑傫 INITIAL_TABLE_SIZE
    ! ł邪Ao^̈g傷Kv邽т2{ɊgB

    integer, parameter:: 	INITIAL_TABLE_SIZE = 4

    ! \
    type(STRING_TABLE_ENTRY), save, pointer::	string_table(:)

    ! ܂܂Ŋt
    integer, save::		table_serial = 0

    integer, parameter::	DISPOSE_HEURISTIC_RAG = 32

contains

    !
    ! === J葱 ===
    !

    !
    !  disposable ֐̎
    !

    subroutine string_set_disposable(str)
	type(STRING), intent(in)::	str
    continue
	string_table(str%id)%nlink = 0
    end subroutine

    !
    !  len ֐̎
    !

    integer function string_len(str) result(result)
	type(STRING), intent(in)::	str
    continue
	if (.not. associated(string_table(str%id)%buffer)) then
	    result = 0
	else
	    result = size(string_table(str%id)%buffer)
	endif
    end function

    !
    !  var_str ֐̎
    !

    type(STRING) function char_to_string(char) result(result)
	character(len=*), intent(in)::	char
    continue
    	!  result = char ̓NJE^1̂Ŗ]܂Ȃ
	result%id = table_set_char(char)
    end function

    !
    !  char ֐̎
    !

    function string_to_char_all(str) result(result)
	type(STRING), intent(in)::	str
	character(len = size(string_table(str%id)%buffer))::	result
    continue
	call char_let_string(result, str)
    end function

    function string_to_char_length(str, length) result(result)
	type(STRING), intent(in)::	str
	integer, intent(in)::			length
	character(len = length)::		result
    continue
	call char_let_string(result, str)
    end function

    !
    ! ̑̎
    !

    subroutine string_let_string(lhs, rhs)
	type(STRING), intent(inout)::	lhs
	type(STRING), intent(in)::	rhs
    continue
!d	write(*, "('#string', i4, ' = string', i4)") lhs%id, rhs%id
	call string_link_decrement(lhs%id)
	lhs%id = rhs%id
	call table_link_increment(rhs%id)
    end subroutine

    subroutine string_let_char(str, char)
	type(STRING), intent(inout)::	str
	character(len=*), intent(in)::		char
    continue
!d	write(*, "('#string', i4, ' = char', a)") str, char
	call string_link_decrement(str%id)
	str%id = table_set_char(char)
	call table_link_increment(str%id)
    end subroutine

    subroutine char_let_string(char, str)
	character(len=*), intent(out)::		char
	type(STRING), intent(in)::	str
    continue
	char = table_get_char(str%id)
    end subroutine

    !
    !  // Zq̎
    !

    integer function string_add_string_id(lhs, rhs) result(result)
	type(STRING), intent(in)::	lhs, rhs
	integer::				lhslen, rhslen
	type(STRING_TABLE_ENTRY), pointer::	result_entry
    continue
!d	write(*, "('#add(entry)', 2i4)") lhs%id, rhs%id
	lhslen = len(lhs)
	rhslen = len(rhs)
	if (lhslen == 0) then
	    result = rhs%id
	    return
	else if (rhslen == 0) then
	    result = lhs%id
	    return
	endif
	result = table_allocate(lhslen + rhslen)
	if (result <= 0) return
	result_entry => string_table(result)
	result_entry%buffer(1: lhslen) = string_table(lhs%id)%buffer(1: lhslen)
	result_entry%buffer(lhslen+1: lhslen+rhslen) = &
	    & string_table(rhs%id)%buffer(1: rhslen)
!d	write(*, "('#add(result)', 3i4)") result, lhs%id, rhs%id
!d	write(*, *) '#<', string_table(result)%buffer, '>=<', &
!d	    & string_table(lhs%id)%buffer(:), '>+<', &
!d	    & string_table(rhs%id)%buffer(:), '>'
    end function

    function string_add_char(str, char) result(result)
	type(STRING)::			result
	type(STRING), intent(in)::	str
	character(len=*), intent(in)::		char
    continue
	result%id = string_add_string_id(str, var_str(char))
    end function

    function string_add_string(lhs, rhs) result(result)
	type(STRING)::			result
	type(STRING), intent(in)::	lhs, rhs
    continue
	result%id = string_add_string_id(lhs, rhs)
    end function

    function char_add_string(char, str) result(result)
	type(STRING)::			result
	character(len=*), intent(in)::		char
	type(STRING), intent(in)::	str
    continue
	result = string_add_string(var_str(char), str)
    end function

    !
    ! ̉Zq == ̎
    !

    logical function string_eq_string(lhs, rhs) result(result)
	type(STRING), intent(in)::	lhs, rhs
    continue
	result = string_cmp_string(lhs, rhs) == CMP_EQUIVALENT
    end function

    logical function string_eq_char(str, char) result(result)
	type(STRING), intent(in)::	str
	character(len=*), intent(in)::		char
    continue
	result = string_eq_string(str, var_str(char))
    end function

    logical function char_eq_string(char, str) result(result)
	character(len=*), intent(in)::		char
	type(STRING), intent(in)::	str
    continue
	result = string_eq_string(var_str(char), str)
    end function

    !
    ! ̉Zq /= ̎
    !

    logical function string_ne_string(lhs, rhs) result(result)
	type(STRING), intent(in)::	lhs, rhs
    continue
	result = .not. string_eq_string(lhs, rhs)
    end function

    logical function string_ne_char(str, char) result(result)
	type(STRING), intent(in)::	str
	character(len=*), intent(in)::		char
    continue
	result = .not. string_eq_char(str, char)
    end function

    logical function char_ne_string(char, str) result(result)
	character(len=*), intent(in)::		char
	type(STRING), intent(in)::	str
    continue
	result = .not. char_eq_string(char, str)
    end function

    !
    ! ̉Zq < ̎
    !

    logical function string_lt_string(lhs, rhs) result(result)
	type(STRING), intent(in)::	lhs, rhs
    continue
	result = string_cmp_string(lhs, rhs) == CMP_LESSTHAN
    end function

    logical function string_lt_char(str, char) result(result)
	type(STRING), intent(in)::	str
	character(len=*), intent(in)::		char
    continue
	result = string_lt_string(str, var_str(char))
    end function

    logical function char_lt_string(char, str) result(result)
	character(len=*), intent(in)::		char
	type(STRING), intent(in)::	str
    continue
	result = string_lt_string(var_str(char), str)
    end function

    !
    ! ̉Zq <= ̎
    !

    logical function string_le_string(lhs, rhs) result(result)
	type(STRING), intent(in)::	lhs, rhs
	integer::				cmp
    continue
	cmp = string_cmp_string(lhs, rhs) 
	result = (cmp == CMP_LESSTHAN) .or. (cmp == CMP_EQUIVALENT)
    end function

    logical function string_le_char(str, char) result(result)
	type(STRING), intent(in)::	str
	character(len=*), intent(in)::		char
    continue
	result = string_le_string(str, var_str(char))
    end function

    logical function char_le_string(char, str) result(result)
	character(len=*), intent(in)::		char
	type(STRING), intent(in)::	str
    continue
	result = string_le_string(var_str(char), str)
    end function

    !
    ! ̉Zq > ̎
    !

    logical function string_gt_string(lhs, rhs) result(result)
	type(STRING), intent(in)::	lhs, rhs
    continue
	result = string_cmp_string(lhs, rhs) == CMP_GREATERTHAN
    end function

    logical function string_gt_char(str, char) result(result)
	type(STRING), intent(in)::	str
	character(len=*), intent(in)::		char
    continue
	result = string_gt_string(str, var_str(char))
    end function

    logical function char_gt_string(char, str) result(result)
	character(len=*), intent(in)::		char
	type(STRING), intent(in)::	str
    continue
	result = string_gt_string(var_str(char), str)
    end function

    !
    ! ̉Zq >= ̎
    !

    logical function string_ge_string(lhs, rhs) result(result)
	type(STRING), intent(in)::	lhs, rhs
	integer::				cmp
    continue
	cmp = string_cmp_string(lhs, rhs) 
	result = (cmp == CMP_GREATERTHAN) .or. (cmp == CMP_EQUIVALENT)
    end function

    logical function string_ge_char(str, char) result(result)
	type(STRING), intent(in)::	str
	character(len=*), intent(in)::		char
    continue
	result = string_ge_string(str, var_str(char))
    end function

    logical function char_ge_string(char, str) result(result)
	character(len=*), intent(in)::		char
	type(STRING), intent(in)::	str
    continue
	result = string_ge_string(var_str(char), str)
    end function

    !
    ! o
    !

    subroutine string_get_default(str, maxlen, iostat)
	type(STRING), intent(out)::	str
	integer, intent(in), optional::		maxlen
	integer, intent(out), optional::	iostat
    continue
	call string_get(-1, str, maxlen, iostat)
    end subroutine

    subroutine string_get(unit, str, maxlen, iostat)
	integer, intent(in)::			unit
	type(STRING), intent(out)::	str
	integer, intent(in), optional::		maxlen
	integer, intent(out), optional::	iostat
	integer::		alreadyread, buflen, nowread, ios, maxsize
	integer, parameter::	BUFFERSIZE = 80
	character(len = BUFFERSIZE)::		buffer
    continue
	if (present(maxlen)) then
	    maxsize = maxlen
	else
	    maxsize = huge(maxsize)
	endif
	alreadyread = 0
	str = ''
	do
	    if (alreadyread >= maxsize) return
	    buflen = min(BUFFERSIZE, maxsize - alreadyread)
	    ! SUPER-UX ΍
	    buffer = ''		
	    ! ǂݎ
	    if (unit >= 0) then
		read(unit=unit, fmt='(A)', advance='NO', &
		    & size=nowread, eor=100, iostat=ios) buffer(1: buflen)
	    else
		read(unit=*, fmt='(A)', advance='NO', &
		    & size=nowread, eor=100, iostat=ios) buffer(1: buflen)
	    endif
	    if (ios /= 0) then
		if (present(iostat)) then
		    iostat = ios
		    return
		else
		    print *, 'get_string: read error ', ios
		    stop
		endif
	    endif
	    ! Ȃ SUPER-UX SX4 Fortran 90 ł͍słȂ
	    if (nowread == 0 .and. len_trim(buffer) /= 0) then
		nowread = len_trim(buffer)
		goto 100
	    endif
	    alreadyread = alreadyread + nowread
	    str = str // buffer(1: nowread)
	enddo
	if (present(iostat)) iostat = 0
	return

	! in case of EOR
100	continue
	str = str // buffer(1: nowread)
	if (present(iostat)) iostat = 0
    end subroutine string_get

    subroutine char_put_default(char, iostat)
	character(len=*), intent(in)::		char
	integer, intent(out), optional::	iostat
    continue
	call char_put(-1, char, iostat)
    end subroutine

    subroutine char_put(unit, char, iostat)
	integer, intent(in)::			unit
	character(len=*), intent(in)::		char
	integer, intent(out), optional::	iostat
	integer:: ios
    continue
	if (unit >= 0) then
	    write(unit=unit, fmt='(A)', advance='NO', iostat=ios) char
	else
	    write(unit=*, fmt='(A)', advance='NO', iostat=ios) char
	endif
	if (present(iostat)) then
	    iostat = ios
	else
	    if (ios /= 0) then
	        print *, 'char_put: write error ', ios
	    endif
	endif
    end subroutine

    subroutine string_put_default(str, iostat)
	type(STRING), intent(in)::	str
	integer, intent(out), optional::	iostat
    continue
	call string_put(-1, str, iostat)
    end subroutine

    subroutine string_put(unit, str, iostat)
	integer, intent(in)::			unit
	type(STRING), intent(in)::	str
	integer, intent(out), optional::	iostat
	integer:: ios
    continue
	if (unit >= 0) then
	    write(unit=unit, fmt='(A)', advance='NO', iostat=ios) char(str)
	else
	    write(unit=*, fmt='(A)', advance='NO', iostat=ios) char(str)
	endif
	if (present(iostat)) then
	    iostat = ios
	else
	    if (ios /= 0) then
	        print *, 'string_put: write error ', ios
	    endif
	endif
    end subroutine

    subroutine char_put_line_default(char, iostat)
	character(len=*), intent(in)::		char
	integer, intent(out), optional::	iostat
    continue
	call char_put_line(-1, char, iostat)
    end subroutine

    subroutine char_put_line(unit, char, iostat)
	integer, intent(in)::			unit
	character(len=*), intent(in)::		char
	integer, intent(out), optional::	iostat
	integer:: ios
    continue
	if (unit >= 0) then
	    write(unit=unit, fmt='(A)', advance='YES', iostat=ios) char
	else
	    write(unit=*, fmt='(A)', advance='YES', iostat=ios) char
	endif
	if (present(iostat)) then
	    iostat = ios
	else
	    if (ios /= 0) then
		print *, 'char_put_line: write error ', ios
	    endif
	endif
    end subroutine

    subroutine string_put_line_default(str, iostat)
	type(STRING), intent(in)::	str
	integer, intent(out), optional::	iostat
    continue
	call string_put_line(-1, str, iostat)
    end subroutine

    subroutine string_put_line(unit, str, iostat)
	integer, intent(in)::			unit
	type(STRING), intent(in)::	str
	integer, intent(out), optional::	iostat
	integer:: ios
    continue
	if (unit >= 0) then
	    write(unit=unit, fmt='(A)', advance='YES', iostat=ios) char(str)
	else
	    write(unit=*, fmt='(A)', advance='YES', iostat=ios) char(str)
	endif
	if (present(iostat)) then
	    iostat = ios
	else
	    if (ios /= 0) then
		print *, 'string_put_line: write error ', ios
	    endif
	endif
    end subroutine

    !
    ! --- gݍ݊֐̏㏑ ---
    !

    !
    ! index ̑p
    !

    function string_index_string(str, substring, back) result(result)
	integer::				result
	type(STRING), intent(in)::	str, substring
	logical, optional::			back
    continue
	result = index(char(str), char(substring), back)
    end function

    function string_index_char(str, substring, back) result(result)
	integer::				result
	type(STRING), intent(in)::	str
	character(len = *), intent(in)::	substring
	logical, optional::			back
    continue
	result = index(char(str), substring, back)
    end function

    function char_index_string(str, substring, back) result(result)
	integer::				result
	character(len = *), intent(in)::	str
	type(STRING), intent(in)::	substring
	logical, optional::			back
    continue
	result = index(str, char(substring), back)
    end function

    !
    ! scan ̑p
    !

    function string_scan_string(str, set, back) result(result)
	integer::				result
	type(STRING), intent(in)::	str, set
	logical, optional::			back
    continue
	result = scan(char(str), char(set), back)
    end function

    function string_scan_char(str, set, back) result(result)
	integer::				result
	type(STRING), intent(in)::	str
	character(len = *), intent(in)::	set
	logical, optional::			back
    continue
	result = scan(char(str), set, back)
    end function

    !
    ! verify ̑p
    !

    function string_verify_string(str, set, back) result(result)
	integer::				result
	type(STRING), intent(in)::	str, set
	logical, optional::			back
    continue
	result = verify(char(str), char(set), back)
    end function

    function string_verify_char(str, set, back) result(result)
	integer::				result
	type(STRING), intent(in)::	str
	character(len = *), intent(in)::	set
	logical, optional::			back
    continue
	result = verify(char(str), set, back)
    end function

    !
    ! === Iɗp ===
    !

    ! rZq
    integer function string_cmp_string(lhs, rhs) result(result)
	type(STRING), intent(in)::	lhs, rhs
	type(STRING_TABLE_ENTRY), pointer::	lhs_entry, rhs_entry
	integer::				lhs_len, rhs_len, i
	character::				lhs_char, rhs_char
    continue
	lhs_entry => string_table(lhs%id)
	rhs_entry => string_table(rhs%id)
	if (.not. associated(lhs_entry%buffer) &
	& .or. .not. associated(rhs_entry%buffer)) then
	    result = CMP_ERROR
	    return
	endif
	result = CMP_EQUIVALENT
	lhs_len = len(lhs)
	rhs_len = len(rhs)
	do, i = 1, max(lhs_len, rhs_len)
	    if (i <= lhs_len) then
		lhs_char = lhs_entry%buffer(i)
	    else
		lhs_char = ' '
	    endif
	    if (i <= rhs_len) then
		rhs_char = rhs_entry%buffer(i)
	    else
		rhs_char = ' '
	    endif
	    if (lhs_char > rhs_char) then
		result = CMP_GREATERTHAN
		return
	    else if (lhs_char < rhs_char) then
		result = CMP_LESSTHAN
		return
	    endif
	enddo
    end function

    !
    ! --- \̑ ---
    !

    ! \̏
    ! ݂̏ꍇQɏI
    function initTable()
	logical::		initTable
	integer::		status, i
    continue
	initTable = .true.
	if (associated(string_table)) return
	allocate(string_table(INITIAL_TABLE_SIZE), stat=status)
	if (status /= 0) then
	    initTable = .false.
	    return
	endif
	! [̕p
	string_table(1)%nlink = 0
	allocate(string_table(1)%buffer(1:0))
	! ȊȌ
	do, i = 2, INITIAL_TABLE_SIZE
	    string_table(i)%nlink = -1
	    nullify(string_table(i)%buffer)
	enddo
    end function

    ! \̍ŏ̎gĂȂT
    integer function firstEmptyId()
	integer::			i
    continue
	do, i = 1, size(string_table)
	    if (string_table(i)%nlink < 0) then
		firstEmptyId = i
		return
	    endif
	enddo
	firstEmptyId = -1
    end function

    ! \̑傫ς
    ! (sĂ񍐂Ȃ)
    subroutine growTable(newsize)
	integer, intent(in)::			newsize
	integer::				copysize, status, i
	type(STRING_TABLE_ENTRY), pointer::	newtable(:)
    continue
	copysize = min(size(string_table), newsize)
	allocate(newtable(newsize), stat=status)
	if (status == 0) then
	    newtable(1:copysize) = string_table(1:copysize)
	    newtable(copysize + 1: newsize)%nlink = -1
	    do, i = copysize + 1, newsize
		nullify(newtable(i)%buffer)
	    enddo
	    deallocate(string_table)
	    string_table => newtable
	endif
    end subroutine

    ! \ id Ԗڂ̎̕QƃJE^1
    subroutine table_link_increment(id)
    	integer, intent(in)::		id
    continue
	if (.not. associated(string_table)) return
	if (id <= 1 .or. id > size(string_table)) return
	if (string_table(id)%nlink < 0) return
	string_table(id)%nlink = string_table(id)%nlink + 1
    end subroutine

    ! \ id Ԗڂ̎̕QƃJE^1
    subroutine string_link_decrement(id)
	integer, intent(in)::		id
    continue
	if (.not. associated(string_table)) return
	if (id <= 1 .or. id > size(string_table)) return
	if (string_table(id)%nlink < 0) return
	string_table(id)%nlink = string_table(id)%nlink - 1
	if (string_table(id)%nlink <= 0) call table_dispose(id)
    end subroutine

    ! \ id Ԗڂ̕Ԃ
    function table_get_char(id) result(result)
	integer, intent(in)::		id
	character(len = size(string_table(id)%buffer))::	result
	integer::			length 
    continue
    	length = size(string_table(id)%buffer)
	result(1: length) = &
	    & transfer(string_table(id)%buffer(1: length), result(1: length))
    end function

    ! id Ԗڂ̈̂̕߂̃p
    subroutine table_dispose(id)
	integer, intent(in)::			id
	integer::				i, enough_old
    continue
	if (id <= 1 .or. id > size(string_table)) return
	call table_dispose_one(id)
	enough_old = string_table(id)%serial - DISPOSE_HEURISTIC_RAG
	do, i = 2, size(string_table)
	    if (string_table(i)%nlink /= 0) cycle
	    if (string_table(i)%serial < enough_old) call table_dispose_one(i)
	enddo
    end subroutine

    subroutine table_dispose_one(id)
	integer, intent(in)::			id
	type(STRING_TABLE_ENTRY), pointer::	cur
    continue
!d	write(*, "('#dispose id', i4, ' <', a, '>')") &
!d	    & id, char(STRING(id))
	cur => string_table(id)
	if (.not. associated(cur%buffer)) return
	deallocate(cur%buffer)
	nullify(cur%buffer)
	cur%nlink = -1
    end subroutine

    ! V̈mۂ length ̗̈tB
    ! ̈ԍ (s畉̒l) Ԃ
    integer function table_allocate(length) result(result)
	integer, intent(in)::			length
	integer::				status
	type(STRING_TABLE_ENTRY), pointer::	cur
    continue
	! `FbNɎs畉̒lԂ
	result = -1
	! `FbN
	if (.not. initTable()) return
	! ̒`FbN
	if (length < 0) return

	! ̒͂ЂƂvȂ
	if (length == 0) then
	    result = 1
	    return
	endif

	! ̒ȂΕ̈m
	result = firstEmptyId()
	if (result < 0) then
	    call growTable(size(string_table) * 2)
	    result = firstEmptyId()
	    if (result < 0) return
	endif
	cur => string_table(result)
	if (associated(cur%buffer)) deallocate(cur%buffer)
	allocate(cur%buffer(1:length), stat=status)
	if (status /= 0) then
	    result = -3
	    return
	endif
	cur%nlink = 0
	table_serial = table_serial + 1
	cur%serial = table_serial
!d	write(*, "('#', 3(a, i4))") 'allocate', length, ' id=', result, &
!d	    & ' serial=', table_serial
    end function

    ! \ɕ char o^ԍԂ
    integer function table_set_char(char) result(result)
	character(len=*), intent(in)::	char
	integer::			length
    continue
	length = len(char)
	result = table_allocate(length)
	if (result < 0) return
	string_table(result)%buffer = transfer(char(1: length), string_table(result)%buffer)
!d	write(*, "('#table_set_char', 2i4, '<', a, '>')") &
!d	    & length, result, char
    end function

end module dc_string
